/*--------------------------------------------------------------------------*/
// Sample Application													
//																			
// Chapter 5																
//																			
// Learn DirectX in 24 Hours												
// by Robert Dunlop															
//																			
// Copyright (C) 1999														
/*--------------------------------------------------------------------------*/

//------ Include Files ------//

#include "stdafx.h"
#define INITGUID
#include <stdio.h>
#include <ddraw.h>
#include <mmsystem.h>

//------ Window Class Information ------//

static char szClass[] = "XmplHr5Class";
static char szCaption[] = "Example - Hour 5";

//------ Define Position Limits ------//

#define MIN_POS				0.0
#define MAX_POS				50000.0

//------ Global Interface Pointers ------//

LPDIRECTDRAW7 lpDD=NULL;
LPDIRECTDRAWSURFACE7 lpDDSPrimary=NULL;
LPDIRECTDRAWSURFACE7 lpDDSBack=NULL;

//------ DirectDraw Surfaces for Object Storage ------//

LPDIRECTDRAWSURFACE7 int_surf=NULL;
LPDIRECTDRAWSURFACE7 back1_surf=NULL;
LPDIRECTDRAWSURFACE7 back2_surf=NULL;
LPDIRECTDRAWSURFACE7 ground_surf=NULL;
LPDIRECTDRAWSURFACE7 light_surf=NULL;

//------ Taxi Layer and Dimensions ------//

LPDIRECTDRAWSURFACE7 taxi_surf=NULL;
WORD taxi_width;
WORD taxi_height;
long taxi_pos=30000.0;

//------ Define Starting Position and Speed ------//

double x_pos=25000.0;			// player position
double move_rate=0.0;			// player motion

//------ Define Number of Paralax Layers ------//

#define	NUM_LAYERS			4

//------ Define Structure for Layers ------//

struct LAYER {
	LPDIRECTDRAWSURFACE7	surf;
	BOOL					blit_flags;
	int						start_x;
	int						start_y;
	WORD					interval;
	double					parallax;
	WORD					width;
	WORD					height;
} layers[NUM_LAYERS];

//------ Storage for Color Key Value ------//

DWORD KeyColor;

//------ Error Return String ------//

const char *ErrStr=NULL;	

//------ Error Messages ------//

const char Err_Reg_Class[]			= "Error Registering Window Class";
const char Err_Create_Win[]			= "Error Creating Window";
const char Err_DirectDrawCreate[]	= "DirectDrawCreate FAILED";
const char Err_Query[]				= "QueryInterface FAILED";
const char Err_Coop[]				= "SetCooperativeLevel FAILED";
const char Err_CreateSurf[]			= "CreateSurface FAILED";
const char Err_GetBack[]			= "Error Retrieving Back Buffer";
const char Err_LoadBMP[]			= "Error Loading Image";
const char Err_DispMode[]			= "Error Setting Display Mode";
const char Err_LoadImage[]			= "Error Loading Image";

//------ Function to Load a Bitmap into a DirectDraw Surface ------//

LPDIRECTDRAWSURFACE7 bitmap_surface(LPCTSTR file_name,RECT *dims=NULL)
{
	HDC hdc;
	HBITMAP bit;
	LPDIRECTDRAWSURFACE7 surf;

	// load the interface bitmap

	bit=(HBITMAP) LoadImage(NULL,file_name,IMAGE_BITMAP,0,0,
								LR_DEFAULTSIZE|LR_LOADFROMFILE);
	if (!bit) 

		// failed to load, return failure to caller

		return NULL;

	// get bitmap dimensions

	BITMAP bitmap;
    GetObject( bit, sizeof(BITMAP), &bitmap );
	int surf_width=bitmap.bmWidth;
	int surf_height=bitmap.bmHeight;

	// create surface

	HRESULT ddrval;
	DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT ;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN; 
	ddsd.dwWidth = surf_width;
	ddsd.dwHeight = surf_height; 

	// attempt to create surface

	ddrval=lpDD->CreateSurface(&ddsd,&surf,NULL);

	// created ok?

	if (ddrval!=DD_OK) {

		// no, release the bitmap and return failure to caller

		DeleteObject(bit);
		return NULL;

	} else {

		// yes, get a DC for the surface

		surf->GetDC(&hdc);

		// generate a compatible DC

		HDC bit_dc=CreateCompatibleDC(hdc);

		// blit the interface to the surface

		SelectObject(bit_dc,bit);
		BitBlt(hdc,0,0,surf_width,surf_height,bit_dc,0,0,SRCCOPY);

		// release the DCs

		surf->ReleaseDC(hdc);
		DeleteDC(bit_dc);

		// save the dimensions if rectangle pointer provided

		if (dims) {
			dims->left=0;
			dims->top=0;
			dims->right=surf_width;
			dims->bottom=surf_height;
		}
	}

	// clear bitmap 

	DeleteObject(bit);

	// return pointer to caller

	return surf;
}

//------ Cleanup Function to Release Objects ------//

#define SafeRelease(x) if (x) { x->Release(); x=NULL;}

void Cleanup()
{
	// release loaded image surfaces

	SafeRelease(back1_surf);
	SafeRelease(back2_surf);
	SafeRelease(int_surf);
	SafeRelease(ground_surf);
	SafeRelease(light_surf);
	SafeRelease(taxi_surf);

	// release DirectDraw interfaces

	SafeRelease(lpDDSPrimary);
	SafeRelease(lpDD);

	// display error if one thrown

	if (ErrStr) {
		MessageBox(NULL, ErrStr, szCaption, MB_OK);
		ErrStr=NULL;
	}
}

//------ Windows Message Handler ------//

LRESULT CALLBACK 
WindowProc(HWND hWnd, unsigned uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        case WM_DESTROY:
			Cleanup();
            PostQuitMessage(0);
            break;

        case WM_KEYDOWN: 
            switch (wParam) 
            { 
                case VK_LEFT: 
                    
                    // Process the LEFT ARROW key. 
                    
					if (move_rate>-300.0)
						move_rate-=60.0;
                    break; 
 
                case VK_RIGHT: 
                    
                    // Process the RIGHT ARROW key. 
                     
					if (move_rate<300.0)
						move_rate+=60.0;
                    break; 
 
				case VK_ESCAPE:

					// exit the program on escape

					DestroyWindow(hWnd);
					break;

                // Process other non-character keystrokes. 
                 
                default: 
                    break; 
            }

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
	return 0L;
}

//------ Function to load object images and set up layers ------//

BOOL load_images()
{
	// load the interface image

	int_surf=bitmap_surface("ntrface.bmp");
	if (!int_surf) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}

	// load the background image

	RECT rct;
	back1_surf=bitmap_surface("back.bmp",&rct);
	if (!back1_surf) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}

	// set up layer record

	layers[0].surf=back1_surf;
	layers[0].blit_flags=DDBLTFAST_WAIT;
	layers[0].start_x=0;
	layers[0].start_y=0;
	layers[0].parallax=0.8;
	layers[0].interval=layers[0].width=rct.right;
	layers[0].height=rct.bottom;

	// load the cityscape layer

	back2_surf=bitmap_surface("back2.bmp",&rct);
	if (!back2_surf) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}

	// set up layer record

	layers[1].surf=back2_surf;
	layers[1].blit_flags=DDBLTFAST_WAIT|DDBLTFAST_SRCCOLORKEY;
	layers[1].start_x=0;
	layers[1].start_y=45;
	layers[1].parallax=1.4;
	layers[1].interval=layers[1].width=rct.right;
	layers[1].height=rct.bottom;

	// load the light layer

	light_surf=bitmap_surface("light.bmp",&rct);
	if (!light_surf) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}

	// set up layer record

	layers[2].surf=light_surf;
	layers[2].blit_flags=DDBLTFAST_WAIT|DDBLTFAST_SRCCOLORKEY;
	layers[2].start_x=0;
	layers[2].start_y=8;
	layers[2].parallax=1.8;
	layers[2].interval=600;
	layers[2].width=rct.right;
	layers[2].height=rct.bottom;

	// load the ground layer

	ground_surf=bitmap_surface("ground.bmp",&rct);
	if (!ground_surf) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}

	// set up layer record

	layers[3].surf=ground_surf;
	layers[3].blit_flags=DDBLTFAST_WAIT;
	layers[3].start_x=0;
	layers[3].start_y=106;
	layers[3].parallax=2.4;
	layers[3].interval=layers[3].width=rct.right;
	layers[3].height=rct.bottom;

	// load the sprite for the taxi

	taxi_surf=bitmap_surface("taxi.bmp",&rct);
	if (!taxi_surf) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}

	// save taxi dimensions

	taxi_width=rct.right;
	taxi_height=rct.bottom;

	// get the pixel format

	DDPIXELFORMAT ddpf;
	ddpf.dwSize=sizeof(ddpf);
	lpDDSPrimary->GetPixelFormat(&ddpf);

	// Determine proper key for pixel format

	KeyColor = ddpf.dwGBitMask;

	// set color keys

	DDCOLORKEY key;
	key.dwColorSpaceLowValue = KeyColor;
	key.dwColorSpaceHighValue = KeyColor;
	int_surf->SetColorKey(DDCKEY_SRCBLT, &key);
	back2_surf->SetColorKey(DDCKEY_SRCBLT, &key);
	light_surf->SetColorKey(DDCKEY_SRCBLT, &key);
	taxi_surf->SetColorKey(DDCKEY_SRCBLT, &key);

	// return success to caller

	return TRUE;
}

//------ Function to Initialize DirectDraw and the Application ------//

static BOOL Init(HINSTANCE hInstance, int nCmdShow)
{
    WNDCLASS                    wc;
    HRESULT                     ddrval;

    // Set up and register window class

    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC) WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(DWORD);
    wc.hInstance = hInstance;
    wc.hIcon = NULL;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szClass;
    if (!RegisterClass(&wc)) {
		ErrStr=Err_Reg_Class;
        return FALSE;
	}

    // Get dimensions of display

    int ScreenWidth = GetSystemMetrics(SM_CXSCREEN);
    int ScreenHeight = GetSystemMetrics(SM_CYSCREEN);

    // Create a window and display

	HWND hWnd;

    hWnd = CreateWindow(szClass,							// class
                        szCaption,							// caption
						WS_VISIBLE|WS_POPUP,				// style 
						0,									// left
						0,									// top
						ScreenWidth,						// width
						ScreenHeight,						// height
                        NULL,								// parent window
                        NULL,								// menu 
                        hInstance,							// instance
                        NULL);								// parms
    if (!hWnd) {
		ErrStr=Err_Create_Win;
        return FALSE;
	}
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

    // Create the main DirectDraw object

    ddrval = DirectDrawCreateEx(NULL, (VOID **) &lpDD, IID_IDirectDraw7, NULL);
    if (ddrval != DD_OK) {
		ErrStr=Err_DirectDrawCreate;
		return FALSE;
	}

    // Set our cooperative level

    ddrval = lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
    if (ddrval != DD_OK) {
		ErrStr=Err_Coop;
		return FALSE;
	}

	// Set the display mode

	ddrval = lpDD->SetDisplayMode( 640, 480, 16, 0, 0);
	if (ddrval !=DD_OK) {
		ErrStr=Err_DispMode;
		return FALSE;
	}

    // Create the primary surface with 1 back buffer

    DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                          DDSCAPS_FLIP | 
                          DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 1;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
	if (ddrval!=DD_OK) {
		ErrStr=Err_CreateSurf;
		return FALSE;
	}

	// Fetch back buffer interface

	DDSCAPS2 ddscaps;
	ZeroMemory(&ddscaps,sizeof(ddscaps));
	ddscaps.dwCaps=DDSCAPS_BACKBUFFER;
	ddrval=lpDDSPrimary->GetAttachedSurface(&ddscaps,&lpDDSBack);
	if (ddrval!=DD_OK) {
		ErrStr=Err_GetBack;
		return FALSE;
	}

	// load the images an set up the layers

	if (!load_images()) 

		// return with error if failed

		return FALSE;

	// return success to caller

	return TRUE;
}

//------ Function to Draw a Frame ------//

void render_frame()
{
	RECT rct;
	long screen_pos;

	// loop through the layers

	for (int i=0;i<NUM_LAYERS;i++) {
	
		// calculate parallax position of layer

		screen_pos=x_pos*layers[i].parallax;

		// set position of first object

		long obj_pos=layers[i].start_x+
					 (screen_pos-layers[i].start_x)
					 /layers[i].width;

		// loop through objects until off right side of screen

		while (obj_pos<screen_pos+640) {

			// is this object on the screen?

			if (obj_pos+layers[i].width>=screen_pos) {

				// yes, calculate the offset to the left edge of the screen

				int screen_x=obj_pos-screen_pos;

				// clip the object if necessary

				rct.left=0;
				rct.top=0;
				rct.right=layers[i].width;
				rct.bottom=layers[i].height;

				// test for clipping

				if (screen_x<0) {
					rct.left-=screen_x;
					screen_x=0;
				} else if (screen_x+rct.right>640) {
					rct.right=640-screen_x;
				}

				if (rct.bottom+layers[i].start_y>480) {
					rct.bottom=480-layers[i].start_y;
				}

				// draw the object to the screen

				lpDDSBack->BltFast(screen_x,layers[i].start_y,layers[i].surf,&rct,layers[i].blit_flags);
			}

			// increment to next object position

			obj_pos+=layers[i].interval;
		}
	}

	// calculate screen position for taxi parallax

	screen_pos=x_pos*3.0;

	// is the taxi on the screen?

	if (taxi_pos+taxi_width>screen_pos&&taxi_pos<screen_pos+640) {

		// yes, calculate the offset to the left edge of the screen

		int screen_x=taxi_pos-screen_pos;

		// clip the object if necessary

		rct.left=0;
		rct.top=0;
		rct.right=taxi_width;
		rct.bottom=taxi_height;
		if (screen_x<0) {
			rct.left-=screen_x;
			screen_x=0;
		} else if (screen_x+rct.right>640) {
			rct.right=640-screen_x;
		}

		// draw the object to the screen

		lpDDSBack->BltFast(screen_x,220,taxi_surf,&rct,DDBLTFAST_WAIT|DDBLTFAST_SRCCOLORKEY);
	}

	// increment the taxi position, but keep withing range so we see it often

	taxi_pos-=3.5;
	if (taxi_pos<screen_pos-5000.0)
		taxi_pos=screen_pos+5000.0;


	// blit the interface to the back buffer with color key

	rct.left=0;
	rct.top=0;
	rct.right=640;
	rct.bottom=480;
	lpDDSBack->BltFast(0,0,int_surf,&rct,DDBLTFAST_WAIT|DDBLTFAST_SRCCOLORKEY);

	// flip to the primary surface

	lpDDSPrimary->Flip(0,DDFLIP_WAIT);
}

//------ Application Loop ------//

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	MSG msg;                // message from queue
	LONGLONG cur_time;       // current time
	BOOL notDone=TRUE;       // flag for thread completion
	DWORD  time_count=16;    // ms per frame, default if no performance counter
	LONGLONG perf_cnt;       // performance timer frequency
	BOOL perf_flag=FALSE;    // flag determining which timer to use
	LONGLONG next_time=0;    // time to render next frame
	LONGLONG last_time=0;	 // time of previous frame
	double time_elapsed;	 // time since previous frame
	double time_scale;		 // scaling factor for time

	// initialize the application, exit on failure

    if (!Init(hInstance, nCmdShow)) {
		Cleanup();
        return FALSE;
	}


	// is there a performance counter available? 

	if (QueryPerformanceFrequency((LARGE_INTEGER *) &perf_cnt)) { 

		// yes, set time_count and timer choice flag 

		perf_flag=TRUE;
		time_count=perf_cnt/60;
		QueryPerformanceCounter((LARGE_INTEGER *) &next_time); 
		time_scale=1.0/perf_cnt;

	} else { 

		// no performance counter, read in using timeGetTime 

		next_time=timeGetTime(); 
		time_scale=0.001;
	} 

	// save time of last frame

	last_time=next_time;

	// run till completed 

	while (notDone) { 

		// is there a message to process? 

		if (PeekMessage( &msg, NULL, 0, 0, PM_REMOVE)) { 

			// yes, is it a quit message? 

			if (msg.message==WM_QUIT) 

				notDone=FALSE; 

			// dispatch the message 

			TranslateMessage(&msg);
			DispatchMessage(&msg); 

		} else {

			// use the appropriate method to get time 
			// and calculate elapsed time since last frame

			if (perf_flag) 
				QueryPerformanceCounter((LARGE_INTEGER *) &cur_time); 
			else 
				cur_time=timeGetTime(); 

			// is it time to render the frame? 

			if (cur_time>next_time) { 

				// calculate elapsed time

				time_elapsed=(cur_time-last_time)*time_scale;

				// save frame time

				last_time=cur_time;
				
				// move the screen position

				x_pos+=move_rate*time_elapsed;
				if (x_pos<MIN_POS) {
					x_pos=MIN_POS;
					move_rate=0;
				}
				if (x_pos>MAX_POS) {
					x_pos=MAX_POS;
					move_rate=0;
				}

				// yes, render the frame 

				render_frame();

				// set time for next frame 

				next_time = cur_time + time_count; 
			}
		}
	}

	// exit returning final message

    return (msg.wParam);
}

